/*
	Copyright (C) 2011 Micha Hersch, University of Lausanne, Swiss Institute of Bioinformatics

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!
 *	\file snake.cpp
 *	\author Micha Hersch
 *	\date
 */
#include "snake.h"
#include "imgGrabber.h"
#include "limits.h"
#define YDIST_THRESHOLD 10

//#define VERBOSE

using namespace std;


Snake::Snake(int h){
  hneb=h;
  neb_size=2*h+1;
  Eint = cvCreateMat(neb_size,neb_size,CV_32FC1);
  Eext = cvCreateMat(neb_size,neb_size,CV_32FC1);
  Etot = cvCreateMat(neb_size,neb_size,CV_32FC1);
  anchored=1;
  size =0;
  valid = 1;
  Eint_weight =1;
}

void Snake::setNeighborhoodSize(int hs){
  cvReleaseMat(&Eint);
  cvReleaseMat(&Eext);
  cvReleaseMat(&Etot);
  hneb=hs;
  neb_size=2*hs+1;
  Eint = cvCreateMat(neb_size,neb_size,CV_32FC1);
  Eext = cvCreateMat(neb_size,neb_size,CV_32FC1);
  Etot = cvCreateMat(neb_size,neb_size,CV_32FC1);
}


Snake::~Snake(){
  cvReleaseMat(&Eint);
  cvReleaseMat(&Eext);
  cvReleaseMat(&Etot);
}



void Snake::updateInternalEnergy(int i){
  Vec2 curr = getPoint(i);
  if(i!=0 && i!= size-1){  //middle points
   Vec2 prev = getPoint(i-1);
    Vec2 next = getPoint(i+1);
   
    Vec2 diff2 = (prev-next)*0.5;
    Vec2 middle =(prev+next)*0.5;
    Vec2 tomiddle;
    Vec2 cand;
    float diff2n = Norm(diff2); 
    for(int j=0;j<neb_size;j++){
      cand.y = curr.y-hneb+j;
      for(int k=0;k<neb_size;k++){
	cand.x = curr.x-hneb+k;
	tomiddle = middle -cand;
	float tomno =  Norm(tomiddle);
	cvmSet(Eint,j,k,2*tomno/(diff2n+tomno));
      }
    }
  }
  else{
    if(!i){//first point (top)
      for(int j=0;j<=hneb;j++){ // 0.5 for the top
	for(int k=0;k<neb_size;k++){
	  cvmSet(Eint,j,k,0.5);
	}
      }
      for(int j=hneb+1;j<neb_size;j++){ //1 for the bottom
	for(int k=0;k<neb_size;k++){
	  cvmSet(Eint,j,k,1);
	}
      }
    }
    else{//last point (bottom)
      if(!anchored){
	for(int j=0;j<hneb;j++){ // 1 for the top
	  for(int k=0;k<neb_size;k++){
	    cvmSet(Eint,j,k,1);
	  }
	}
	for(int j=hneb;j<neb_size;j++){ //0.3 for the bottom
	  for(int k=0;k<neb_size;k++){
	    cvmSet(Eint,j,k,0.3);
	  }
	}
      }
      else{//can only move horizonally
	for(int j=0;j<neb_size;j++){ 
	  for(int k=0;k<neb_size;k++){
	    cvmSet(Eint,j,k,1000);
	  }
	}
	int dx = sign(getPoint(i-1).x-curr.x); 
	for(int k=0;k<neb_size;k++){
	  cvmSet(Eint,hneb,k,dx*sign(hneb-k)*0.01+0.5);
	}
      }
    }
  }
}



void Snake::updateExternalEnergy(int i, IplImage *img){
  Vec2 curr = getPoint(i);
  CvRect rect;
  double mmin,mmax;
  cvResetImageROI(img);//needed ??
  if(curr.x>=hneb && curr.y>=hneb && //whole neighborhood inside image
     img->width>curr.x+hneb &&img->height>curr.y+hneb){
    rect = cvRect(curr.x-hneb,curr.y-hneb,neb_size,neb_size);
    cvSetImageROI(img,rect);
    cvConvert(img,Eext);  
    cvMinMaxLoc(Eext,&mmin,&mmax);
  }
  else{// part of the neighborhood is outside of image TO TEST
     int nx=curr.x>=hneb?0:hneb-curr.x;
     int ny=curr.y>=hneb?0:hneb-curr.y;
     int sx=img->width-curr.x>hneb? neb_size-nx : hneb+img->width-curr.x;
     int sy=img->height-curr.y>hneb? neb_size-ny : hneb+img->height-curr.y;
     
     CvRect erect = cvRect(nx,ny,sx,sy);
     CvRect irect = cvRect(curr.x-hneb+nx,curr.y-hneb+ny,sx,sy);
     
     cvSet(Eext,cvRealScalar(1000));     
     CvMat *parEext = cvCreateMatHeader(sy,sx,CV_32FC1);
     CvMat *parEext2 = cvGetSubRect(Eext,parEext, erect);
     cvSetImageROI(img,irect);
     cvConvert(img,parEext2);  
     cvMinMaxLoc(parEext2,&mmin,&mmax);
     cvReleaseMatHeader(&parEext);
     //    cvReleaseMatHeader(&parEext2);
}
  cvScale(Eext,Eext,1/mmax);    
  cvResetImageROI(img);
     
}

void Snake::printEnergies(){
  cout<<"internal "<<endl;
  for(int i=0;i<neb_size;i++){
    for(int j=0;j<neb_size;j++){
      cout<<cvmGet(Eint,i,j)<<" ";
    }
    cout<<endl;
  }
  cout<<endl<<"external "<<endl;
  for(int i=0;i<neb_size;i++){
    for(int j=0;j<neb_size;j++){
      cout<<cvmGet(Eext,i,j)<<" ";
    }
    cout<<endl;
  }

  cout<<endl<<"total "<<endl;
  for(int i=0;i<neb_size;i++){
    for(int j=0;j<neb_size;j++){
      cout<<cvmGet(Etot,i,j)<<" ";
    }
    cout<<endl;
  }

}

void Snake::updatePoint(int p, IplImage *img){
#ifdef RANDOM_WEIGHT
 float fac=RND(2*Eint_weight);
#else
 float fac=Eint_weight;
#endif
  double emin,emax;
  Vec2 curr = getPoint(p);
  CvPoint minpos;
  updateInternalEnergy(p);
  updateExternalEnergy(p,img);
  cvAddWeighted(Eext,1,Eint,fac,0,Etot);
  cvMinMaxLoc(Etot,&emin,&emax,&minpos);
  points[p] = points[p]+Vec2(minpos)-Vec2(hneb,hneb);
}


void Snake::updateSnake(IplImage *img,int nbit, int eyecup){
  int a = (eyecup!=0);
  int b = (eyecup<0);//don't update the last point
  for(int k=0;k<nbit;k++){
    u32 *indices = randPerm(size-a-b);//random order for points update
    for(int i=0;i<size-a-b;i++){
      updatePoint((int)indices[i]+a,img);
    }
    delete[] indices;
  }
}


void Snake::updateTip(IplImage *img, int nbit,int n){
  int m = max(n,size);
  for(int k=0;k<nbit;k++){
    for(int i=0;i<m;i++){
      updatePoint(i,img);
    } 
  }
}

void Snake::downFromTip(IplImage *img, float angle,float len){
  float a = RAD(angle/2);
  float step = len/(float)size;
  for(int i=1;i<size;i++){
    points[i] = points[0] + Vec2(((i%2)*2-1)*sin(a),cos(a))*i*step;
    points[i].x = TRIM(0, points[i].x,img->width);
    points[i].y= TRIM(0, points[i].y,img->height);
  }
  // cout<<points[0]<<" "<<points[1]<<" "<<points[2]<<endl;
  updateSnake(img,10,1);
}   

int Snake::init(IplImage *img,int nb_el, CvRect *rect,float canny,
		int ground,int threshold){
  CvSeq *seq,*maxseq;
  CvPoint *p;
  IplImage *edges,*bwim;
  int ox,oy;
  //  cot.init(img);//initializing the cotyledon
  if(rect){
    edges=cvCreateImage(cvSize(rect->width,rect->height),8,1);
    ox = rect->x;
    oy = rect->y;
    cvSetImageROI(img,*rect);
  
  }
  else{// all images
    edges=cvCloneImage(img);
    ox=oy=0;
  }

  if(threshold){
    bwim=cvCreateImage(cvSize(rect->width,rect->height),8,1);
    cvThreshold(img,bwim,threshold,255,CV_THRESH_BINARY);
  }
  else{
    bwim = img;
    cvConvert(img,edges);
  }  
  cvCanny(bwim,edges,canny*0.6*255,canny*255);
#ifdef VERBOSE
   cvNamedWindow("edges", CV_WINDOW_AUTOSIZE); 
   cvShowImage("edges",edges);
   cvWaitKey();
#endif
  CvMemStorage *storage= cvCreateMemStorage();
  CvContourScanner scan = 
    cvStartFindContours(edges,storage,sizeof(CvContour),
			CV_RETR_LIST,CV_CHAIN_APPROX_NONE,cvPoint(ox,oy));
  // looking for the biggest contour
  seq = cvFindNextContour(scan); 
  maxseq = seq;
  // cout<<ground<<endl;
  if(ground && seq){// we have the ground value
   
    int minydist = findMinVertDist(seq,ground);
    minydist =max(0,minydist-YDIST_THRESHOLD);
    while((seq = cvFindNextContour(scan))){
      int ydist = findMinVertDist(seq,ground);
      ydist = max(0,ydist-YDIST_THRESHOLD);
      if(ydist<minydist || (// closer enough
			    ydist==minydist && seq->total > maxseq->total)){
	maxseq = seq;
	minydist=ydist;
      }
      //      cout<<"total "<<seq->total<<endl;
    }
  }
  else{// we don't have the ground value -> take the biggest
    while((seq = cvFindNextContour(scan))){
      if(seq->total > maxseq->total){
	maxseq = seq;
      }
    }
  }
  if(maxseq){
    empty();
    cvSeqSort( maxseq, cmp_func, 0);
    if(maxseq->total >= nb_el){
      int step= max(1,maxseq->total/nb_el);//so that step>0    
      for(int i=0;i<maxseq->total;i+=step){
	p = CV_GET_SEQ_ELEM(CvPoint,maxseq,i);
	addPoint(Vec2(*p));
      }
    }
    else{
      int step= nb_el/maxseq->total;//so that step>0 
      int sup = nb_el%maxseq->total;
      for(int i=0;i<maxseq->total;i++){
	p = CV_GET_SEQ_ELEM(CvPoint,maxseq,i);
	for(int j=0;j<step;j++){
	  addPoint(Vec2(*p));
	}
	if(RND(maxseq->total)<sup){
	  addPoint(Vec2(*p));
	}
      }
    }
  }

 
  cvResetImageROI(img);
  if(threshold){IMKILL(bwim);}
  IMKILL(edges);  
  cvReleaseMemStorage( &storage );
  return size;
}

void Snake::drop(int x,int y, int si,int spread){
  size= max(0,min(si,MAX_SNAKE_SIZE));
  for(int i=0;i<size;i++){
    Vec2 v(x+RND(spread)-spread/2.0,y+RND(spread)-spread/2.0);
    points[i] = v.round();
  }
}
	

//assumes that points are sorted according to y
void Snake::anchorToGround(int ground){
  anchored=1;
  for(int i=size-1;i>=0;i--){
    if(points[i].y>ground){
      size--;
    }
    else{
      break;
    }
  }
  points[size-1].y = ground;
  getBase().setPosition(getLastPoint());
}


int Snake::findClosestPoint(Vec2 pos, int *sqdist){
  float  mindist = (getPoint(0)-pos).lengthSquared();
  int mini= 0;
  for(int i=1;i<size;i++){
    float dist = (getPoint(i)-pos).lengthSquared();
    if(mindist>=dist){
      mindist=dist;
      mini=i;
    }
  }
  *sqdist = mindist;
  return mini;
}

void Snake::drawLength(IplImage *disp,float len, CvScalar col){
  if(!size)return;
  float sum= len;
  float l=0;
  int i= 0;
  while(sum>0){
    if(i)cvLine(disp,points[i-1].to2d(),points[i].to2d(),col,1);
    l = (points[i]-points[i+1]).length();
    sum -= l;
    i++;
  }
  Vec2 p1 =points[i-1]+(points[i]-points[i-1])*((sum+l)/l);
  cvLine(disp,points[i-1].to2d(),p1.to2d(),col,1);
}


// no check if the length is actually possible
float Snake::getOrientation(float len)const{
  if(!size)return 1000;// error code
  float sum= len;
  float l=0;
  int i= 0;
  while(sum>0){
    l = (points[i]-points[i+1]).length();
    sum -= l;
    i++;
  }
  Vec2 p1 =points[i-1]+(points[i]-points[i-1])*((sum+l)/l);
  Vec2 p = points[0]-p1;
  return DEG(atan2(-p.x,-p.y)); // vertical corresponds to zero angle
}

// returns length in pixel
float Snake::getLength()const{
  float sum = 0;
  for(int i=1;i<size;i++){
    sum+= (points[i-1]-points[i]).length();
  }
  return sum;
}

void Snake::printPoints(ostream& out)const{
  out<<size<<" ";
  for(int i=0;i<size;i++){
    out<<points[i]<<" ";
  }
}


int findMinVertDist(CvSeq *seq, int y0){
  int mindist,dist;
  if (seq->total<=2){return INT_MAX;}
  mindist = (CV_GET_SEQ_ELEM(CvPoint,seq,0))->y -y0;
  mindist = abs(mindist);
  if(mindist<=YDIST_THRESHOLD){return mindist;}
  for(int i=1;i<seq->total;i++){
   dist= (CV_GET_SEQ_ELEM(CvPoint,seq,i))->y-y0;
   dist =abs(dist);
    if(dist<=YDIST_THRESHOLD){return dist;}
    if(dist<mindist){mindist=dist;}
  }
  return mindist;
}





